

/*

Interface the FTDI dll  "FTD2XX" in VIsual Studio C++

05-Mar-2023  Created


*/




#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>



/* ----------- FTDI Dll stuff -------------- */

#define FTDIDLL "FTD2XX.dll" 

HINSTANCE hftdiDLL;

// FTDI constants
//
const uint32_t FT_LIST_NUMBER_ONLY = 0x80000000;
const uint32_t FT_LIST_BY_INDEX = 0x40000000;
const uint32_t FT_LIST_ALL = 0x20000000;

const uint32_t FT_OPEN_BY_SERIAL_NUMBER = 1;
const uint32_t FT_OPEN_BY_DESCRIPTION = 2;
const uint32_t FT_OPEN_BY_LOCATION = 4;

const uint32_t FT_OK = 0;


// Functions from the FTDI chip dll
//
typedef INT32  (__cdecl *FT_ListDevices) (void*, void*, int);
typedef void  (__cdecl *FT_CreateDeviceInfoList) (INT32* numDevices);
typedef void (__cdecl  *FT_GetDeviceInfoDetail) (uint32_t temp, uint32_t* Flags, uint32_t* deviceType, uint32_t* id, uint32_t* locID, char* pcSerialNumber, char* pcDescription, uint64_t* Handle);
typedef INT32 (__cdecl *FT_OpenEx) (void* serialNumber, uint32_t mode, uint64_t* Handle);
typedef INT32 (__cdecl  *FT_GetBitMode) (uint64_t Handle, uint8_t* bitMode);
typedef INT32(__cdecl *FT_SetBitMode) (uint64_t Handle, uint8_t mask, uint8_t mode);
typedef INT32 (__cdecl *FT_Write) (uint64_t Handle, void* bRelayStates, uint32_t buflength, uint32_t* tempInt);
typedef INT32 (__cdecl *FT_Close) (uint64_t Handle);
typedef INT32(__cdecl *FT_Read) (uint64_t Handle, void* buffer, uint32_t bytesToRead, uint32_t* bytesRead);






FT_ListDevices                     lpfnFT_ListDevices; // Function pointer to FT_ListDevices
FT_CreateDeviceInfoList            lpfnFT_CreateDeviceInfoList;
FT_GetDeviceInfoDetail             lpfnFT_GetDeviceInfoDetail;
FT_OpenEx                          lpfnFT_OpenEx;
FT_GetBitMode                      lpfnFT_GetBitMode;
FT_SetBitMode                      lpfnFT_SetBitMode;
FT_Write                           lpfnFT_Write;
FT_Read                            lpfnFT_Read;
FT_Close                           lpfnFT_Close;


void debug(const char* string)
{
    printf(string);
}



/*
Search for attached TCTEC Super4 Relay boards.
The stringArray will be filled with the names of the relay boards.
It must point to a 2 dimensional char (8bit) array of dimensions [n][20]
where n is the number of boards expected.

Returns:
         The number of relay boards found or 0 if none found.

*/
INT32 ListRelays(char stringArray[][20])
{
    int32_t numDevices;
    uint32_t temp;
    uint32_t Flags;
    uint32_t deviceType;
    uint32_t id;
    uint32_t locID;
    char pcSerialNumber[100];
    char pcDescription[200];
    uint64_t Handle;

    int relayCount = 0;

    // get details of connected FTDI devices
    //	
    (lpfnFT_CreateDeviceInfoList)(&numDevices);


    // list all the devices with the description : "TCTEC USB RELAY"
    //

    for (temp = 0; temp < numDevices; temp++)
    {
        (lpfnFT_GetDeviceInfoDetail)(temp, &Flags, &deviceType, &id, &locID, pcSerialNumber, pcDescription, &Handle);

        if (strstr(pcDescription, "TCTEC USB RELAY") != NULL)
        {
            // found one
            //
            strncpy_s(stringArray[relayCount], pcSerialNumber, 19);
            relayCount += 1;
        }

    }

    return(relayCount);
}

/*
Set the relay states.
The first 4 bits of the 'mask' determines which relays will be affected:
    1 = relay will be affected
    0 = leave alone (don't change)

'set' determines whether the relay will be switched on or switched off.
1 = relay on
0 = relay off.

Returns:
        The state of the relays after the action.
        0 to 15:
        first 4 bits represent the relay states.
        1 = relay is on.
        0 = relay is off.

        A negative number indicates an error:

        -1 invalid relay board string (board not found)
        -2 DLL function not loaded (FTD2XX.DLL not found)
        -3 Invalid bit mask (must be 0 to 15)
        -4 Undefined error

*/
INT32 SetRelays(char* relayBoard, uint8_t mask, uint8_t set)
{

    uint32_t tempInt;
    uint64_t Handle;
    uint8_t bRelayStates = 0;
    uint8_t bitMode;


    // open a device and set to outputs, may already be done.
    //		
    tempInt = (lpfnFT_OpenEx)((void*)relayBoard, FT_OPEN_BY_SERIAL_NUMBER, &Handle);
    if (tempInt != FT_OK)
    {
        return(-1);
    }

    tempInt = (uint32_t) (lpfnFT_GetBitMode)(Handle, &bitMode);
    if (tempInt != FT_OK)
    {
        return(-4);
    }

    // The bit mode only needs to be set once after first plugged in
    //		
    if ((bitMode >> 4) > 0)
    {
        // bit mode has not yet been set, so set it
        //		
        // set the bit mode (all inputs)
        //

        tempInt = (lpfnFT_SetBitMode)(Handle, 0xF0, 0x01);
        if (tempInt != FT_OK)
        {
            return(-4);
        }

        // set all to high (off)
        //		
        bRelayStates = 15;

        // write twice, as there is buffering
        //
        (lpfnFT_Write)(Handle, (void*)&bRelayStates, 1, &tempInt);
        (lpfnFT_Write)(Handle, (void*)&bRelayStates, 1, &tempInt);

        // Set up asyncronous bit bang mode all outputs		
        //				
        (lpfnFT_SetBitMode)(Handle, 0xFF, 0x01);
    }

    // read the current state of the bits (relays)
    //	
    (lpfnFT_Read)(Handle, (void*)&bRelayStates, 1, (uint32_t*)&tempInt);


    // invert states, because 0 = on, 1 = off
    //	
    bRelayStates = ~bRelayStates;

    // if turning on, then OR with current state
    //
    if (set)
    {
        bRelayStates |= mask;
    }
    else
    {
        bRelayStates &= ~(mask); // turn off
    }
    // set the relays, invert before writing out. (0 = on, 1 = off)
    //
    bRelayStates = ~bRelayStates;


    // write twice, as there is buffering
    //
    (lpfnFT_Write)(Handle, (void*)&bRelayStates, 1, &tempInt);
     (lpfnFT_Write)(Handle, (void*)&bRelayStates, 1, &tempInt);


    tempInt = (lpfnFT_Close)(Handle);

    if (tempInt != FT_OK)
    {
        return(-4);
    }

    // read the current state of the bits (relays)
    //	
    (lpfnFT_Read)(Handle, (void*)&bRelayStates, 1, (uint32_t*)&tempInt);

    bRelayStates = (~bRelayStates & 15);
    return((int)bRelayStates);
}



//BOOL APIENTRY DllMain(HINSTANCE hInst     /* Library instance handle. */,
//    DWORD reason        /* Reason this function is being called. */,
//    LPVOID reserved     /* Not used. */)
//{
//    switch (reason)
//    {
//    case DLL_PROCESS_ATTACH:
//        return (setupLibrary());
//        break;

//case DLL_PROCESS_DETACH:
//        if (hftdiDLL)
//        {
//            FreeLibrary(hftdiDLL);
//            //   debug("Detached from ftdi dll") ;  
//        }
//        break;

//    case DLL_THREAD_ATTACH:
//        break;

//    case DLL_THREAD_DETACH:
//        break;
//    }

    ///* Returns TRUE on success, FALSE on failure */
//    return (TRUE);
//}


/*
Called once when a process first loads this dll
Link to FTDI DLL and get function pointers.
*/
BOOL setupLibrary()
{
    //debug ("Setup Library") ;

   // Load the FTD2XX.DLL
   //   
   // from example here:
   //https://learn.microsoft.com/en-us/windows/win32/dlls/using-run-time-dynamic-linking

    
    hftdiDLL = LoadLibraryW(TEXT("FTD2XX.dll"));




    if (hftdiDLL != NULL)
    {
        debug("Success. linked to FTD2XX") ;                 

        // Get pointers to the functions in the DLL
        //
        lpfnFT_ListDevices = (FT_ListDevices) GetProcAddress(hftdiDLL, "FT_ListDevices");
        lpfnFT_CreateDeviceInfoList = (FT_CreateDeviceInfoList)GetProcAddress(hftdiDLL, "FT_CreateDeviceInfoList");
        lpfnFT_GetDeviceInfoDetail = (FT_GetDeviceInfoDetail)GetProcAddress(hftdiDLL, "FT_GetDeviceInfoDetail");
        lpfnFT_OpenEx = (FT_OpenEx)GetProcAddress(hftdiDLL, "FT_OpenEx");
        lpfnFT_GetBitMode = (FT_GetBitMode)GetProcAddress(hftdiDLL, "FT_GetBitMode");
        lpfnFT_SetBitMode = (FT_SetBitMode)GetProcAddress(hftdiDLL, "FT_SetBitMode");
        lpfnFT_Write = (FT_Write)GetProcAddress(hftdiDLL, "FT_Write");
        lpfnFT_Read = (FT_Read)GetProcAddress(hftdiDLL, "FT_Read");
        lpfnFT_Close = (FT_Close)GetProcAddress(hftdiDLL, "FT_Close");

        if (!lpfnFT_ListDevices || !lpfnFT_CreateDeviceInfoList || !lpfnFT_GetDeviceInfoDetail || !lpfnFT_OpenEx || !lpfnFT_GetBitMode || !lpfnFT_SetBitMode || !lpfnFT_Write || !lpfnFT_Read)
        {
            // Couldn't find the functions, so exit
            //
            FreeLibrary(hftdiDLL);
            debug("Failed. linking to functions.") ; 
            return (FALSE);
        }
        return(TRUE);  //  success   
    }
    else
    {
        return(FALSE);
    }
}
